@@ -1,21 +1,21 @@
 /*
- * Copyright (C) 2000  Internet Software Consortium.
- * 
+ * Copyright (C) 2000, 2001  Internet Software Consortium.
+ *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
- * 
- * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
- * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
- * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
- * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
- * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
- * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
- * SOFTWARE.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM
+ * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
+ * INTERNET SOFTWARE CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT,
+ * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
+ * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
+ * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
+ * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id: nsupdate.c,v 1.8.2.7 2000/09/15 23:47:14 gson Exp $ */
+/* $Id: nsupdate.c,v 1.75.2.1 2001/01/09 22:32:54 bwelling Exp $ */
 
 #include <config.h>
 
@@ -26,17 +26,22 @@
 #include <stdlib.h>
 #include <unistd.h>
 
+#if (!(defined(HAVE_ADDRINFO) && defined(HAVE_GETADDRINFO)))
+extern int h_errno;
+#endif
+
+#include <isc/app.h>
 #include <isc/base64.h>
 #include <isc/buffer.h>
-#include <isc/condition.h>
 #include <isc/commandline.h>
 #include <isc/entropy.h>
+#include <isc/event.h>
 #include <isc/lex.h>
 #include <isc/mem.h>
-#include <isc/mutex.h>
 #include <isc/region.h>
 #include <isc/sockaddr.h>
 #include <isc/socket.h>
+#include <isc/stdio.h>
 #include <isc/string.h>
 #include <isc/task.h>
 #include <isc/timer.h>
@@ -46,6 +51,7 @@
 #include <dns/callbacks.h>
 #include <dns/dispatch.h>
 #include <dns/events.h>
+#include <dns/fixedname.h>
 #include <dns/message.h>
 #include <dns/name.h>
 #include <dns/rdata.h>
@@ -63,13 +69,14 @@
 #include <lwres/lwres.h>
 #include <lwres/net.h>
 
-#define MXNAME 256
-#define MAXPNAME 1025
-#define MAXCMD 1024
+#define MAXCMD (4 * 1024)
+#define INITDATA (32 * 1024)
+#define MAXDATA (64 * 1024)
 #define NAMEBUF 512
 #define WORDLEN 512
-#define PACKETSIZE 2048
-#define MSGTEXT 4096
+#define PACKETSIZE ((64 * 1024) - 1)
+#define INITTEXT (2 * 1024)
+#define MAXTEXT (128 * 1024)
 #define FIND_TIMEOUT 5
 #define TTL_MAX 2147483647	/* Maximum signed 32 bit integer. */
 
@@ -77,15 +84,15 @@
 
 #define RESOLV_CONF "/etc/resolv.conf"
 
-static isc_boolean_t busy = ISC_FALSE;
 static isc_boolean_t debugging = ISC_FALSE, ddebugging = ISC_FALSE;
+static isc_boolean_t memdebugging = ISC_FALSE;
+static isc_boolean_t have_ipv4 = ISC_FALSE;
 static isc_boolean_t have_ipv6 = ISC_FALSE;
 static isc_boolean_t is_dst_up = ISC_FALSE;
 static isc_boolean_t usevc = ISC_FALSE;
-static isc_mutex_t lock;
-static isc_condition_t cond;
 static isc_taskmgr_t *taskmgr = NULL;
 static isc_task_t *global_task = NULL;
+static isc_event_t *global_event = NULL;
 static isc_mem_t *mctx = NULL;
 static dns_dispatchmgr_t *dispatchmgr = NULL;
 static dns_requestmgr_t *requestmgr = NULL;
@@ -99,15 +106,19 @@ static dns_name_t *origin; /* Points to one of above, or dns_rootname */
 static dns_fixedname_t fuserzone;
 static dns_name_t *userzone = NULL;
 static dns_tsigkey_t *key = NULL;
-static dns_tsig_keyring_t *keyring = NULL;
 static lwres_context_t *lwctx = NULL;
 static lwres_conf_t *lwconf;
 static isc_sockaddr_t *servers;
 static int ns_inuse = 0;
 static int ns_total = 0;
 static isc_sockaddr_t *userserver = NULL;
+static isc_sockaddr_t *localaddr = NULL;
 static char *keystr = NULL, *keyfile = NULL;
 static isc_entropy_t *entp = NULL;
+static isc_boolean_t shuttingdown = ISC_FALSE;
+static FILE *input;
+static isc_boolean_t interactive = ISC_TRUE;
+static isc_boolean_t seenerror = ISC_FALSE;
 
 typedef struct nsu_requestinfo {
 	dns_message_t *msg;
@@ -115,8 +126,8 @@ typedef struct nsu_requestinfo {
 } nsu_requestinfo_t;
 
 static void
-sendrequest(isc_sockaddr_t *address, dns_message_t *msg,
-	    dns_request_t **request);
+sendrequest(isc_sockaddr_t *srcaddr, isc_sockaddr_t *destaddr,
+	    dns_message_t *msg, dns_request_t **request);
 
 #define STATUS_MORE	(isc_uint16_t)0
 #define STATUS_SEND	(isc_uint16_t)1
@@ -127,7 +138,7 @@ static void
 fatal(const char *format, ...) {
 	va_list args;
 
-	va_start(args, format);	
+	va_start(args, format);
 	vfprintf(stderr, format, args);
 	va_end(args);
 	fprintf(stderr, "\n");
@@ -139,7 +150,7 @@ debug(const char *format, ...) {
 	va_list args;
 
 	if (debugging) {
-		va_start(args, format);	
+		va_start(args, format);
 		vfprintf(stderr, format, args);
 		va_end(args);
 		fprintf(stderr, "\n");
@@ -151,7 +162,7 @@ ddebug(const char *format, ...) {
 	va_list args;
 
 	if (ddebugging) {
-		va_start(args, format);	
+		va_start(args, format);
 		vfprintf(stderr, format, args);
 		va_end(args);
 		fprintf(stderr, "\n");
@@ -247,17 +258,12 @@ setup_key(void) {
 	dns_fixedname_t fkeyname;
 	dns_name_t *keyname;
 
-	result = dns_tsigkeyring_create(mctx, &keyring);
-	check_result(result, "dns_tsigkeyringcreate");
-
 	dns_fixedname_init(&fkeyname);
 	keyname = dns_fixedname_name(&fkeyname);
 
 	if (keystr != NULL) {
 		isc_buffer_t keynamesrc;
 		char *secretstr;
-		isc_buffer_t secretsrc;
-		isc_lex_t *lex = NULL;
 		char *s;
 
 		debug("Creating key...");
@@ -280,27 +286,16 @@ setup_key(void) {
 		if (secret == NULL)
 			fatal("out of memory");
 
-		isc_buffer_init(&secretsrc, secretstr, strlen(secretstr));
-		isc_buffer_add(&secretsrc, strlen(secretstr));
-
 		isc_buffer_init(&secretbuf, secret, secretlen);
-
-		result = isc_lex_create(mctx, strlen(secretstr), &lex);
-		check_result(result, "isc_lex_create");
-		result = isc_lex_openbuffer(lex, &secretsrc);
-		check_result(result, "isc_lex_openbuffer");
-		result = isc_base64_tobuffer(lex, &secretbuf, -1);
+		result = isc_base64_decodestring(mctx, secretstr, &secretbuf);
 		if (result != ISC_R_SUCCESS) {
 			fprintf(stderr, "Couldn't create key from %s: %s\n",
 				keystr, isc_result_totext(result));
-			isc_lex_close(lex);
-			isc_lex_destroy(&lex);
 			goto failure;
 		}
+
 		secretlen = isc_buffer_usedlength(&secretbuf);
 		debug("close");
-		isc_lex_close(lex);
-		isc_lex_destroy(&lex);
 	} else {
 		dst_key_t *dstkey = NULL;
 
@@ -328,11 +323,11 @@ setup_key(void) {
 		dst_key_free(&dstkey);
 				    
 	}
-		
+
 	debug("keycreate");
 	result = dns_tsigkey_create(keyname, dns_tsig_hmacmd5_name,
 				    secret, secretlen, ISC_TRUE, NULL, 0, 0,
-				    mctx, keyring, &key);
+				    mctx, NULL, &key);
 	if (result != ISC_R_SUCCESS) {
 		char *str;
 		if (keystr != NULL)
@@ -349,7 +344,17 @@ setup_key(void) {
 
 	if (secret != NULL)
 		isc_mem_free(mctx, secret);
-	dns_tsigkeyring_destroy(&keyring);
+}
+
+static void
+shutdown_program(isc_task_t *task, isc_event_t *event) {
+	REQUIRE(task == global_task);
+	UNUSED(task);
+
+	ddebug("shutdown_program()");
+	isc_event_free(&event);
+	isc_task_detach(&global_task);
+	shuttingdown = ISC_TRUE;
 }
 
 static void
@@ -363,22 +368,19 @@ setup_system(void) {
 
 	ddebug("setup_system()");
 
-	/*
-	 * Warning: This is not particularly good randomness.  We'll
-	 * just use random() now for getting id values, but doing so
-	 * does NOT insure that id's can't be guessed.
-	 *
-	 * XXX Shouldn't random() be called somewhere if this is here?
-	 */
-	srandom(getpid() + (int)&setup_system);
+	dns_result_register();
 
 	result = isc_net_probeipv4();
-	check_result(result, "isc_net_probeipv4");
+	if (result == ISC_R_SUCCESS)
+		have_ipv4 = ISC_TRUE;
 
 	result = isc_net_probeipv6();
 	if (result == ISC_R_SUCCESS)
 		have_ipv6 = ISC_TRUE;
 
+	if (!have_ipv4 && !have_ipv6)
+		fatal("couldn't find either IPv4 or IPv6");
+
 	result = isc_mem_create(0, 0, &mctx);
 	check_result(result, "isc_mem_create");
 
@@ -427,6 +429,9 @@ setup_system(void) {
 	result = isc_task_create(taskmgr, 0, &global_task);
 	check_result(result, "isc_task_create");
 
+	result = isc_task_onshutdown(global_task, shutdown_program, NULL);
+	check_result(result, "isc_task_onshutdown");
+
 	result = isc_entropy_create(mctx, &entp);
 	check_result(result, "isc_entropy_create");
 
@@ -449,14 +454,17 @@ setup_system(void) {
 		check_result(result, "dns_dispatch_getudp (v6)");
 	}
 
-	attrs = DNS_DISPATCHATTR_UDP;
-	attrs |= DNS_DISPATCHATTR_MAKEQUERY;
-	attrs |= DNS_DISPATCHATTR_IPV4;
-	isc_sockaddr_any(&bind_any);
-	result = dns_dispatch_getudp(dispatchmgr, socketmgr, taskmgr,
-				     &bind_any, PACKETSIZE, 4, 2, 3, 5,
-				     attrs, attrmask, &dispatchv4);
-	check_result(result, "dns_dispatch_getudp (v4)");
+	if (have_ipv4) {
+		attrs = DNS_DISPATCHATTR_UDP;
+		attrs |= DNS_DISPATCHATTR_MAKEQUERY;
+		attrs |= DNS_DISPATCHATTR_IPV4;
+		isc_sockaddr_any(&bind_any);
+		result = dns_dispatch_getudp(dispatchmgr, socketmgr, taskmgr,
+					     &bind_any, PACKETSIZE,
+					     4, 2, 3, 5,
+					     attrs, attrmask, &dispatchv4);
+		check_result(result, "dns_dispatch_getudp (v4)");
+	}
 
 	result = dns_requestmgr_create(mctx, timermgr,
 				       socketmgr, taskmgr, dispatchmgr,
@@ -483,21 +491,21 @@ setup_system(void) {
 
 static void
 get_address(char *host, in_port_t port, isc_sockaddr_t *sockaddr) {
-        struct in_addr in4;
-        struct in6_addr in6;
+	struct in_addr in4;
+	struct in6_addr in6;
 #if defined(HAVE_ADDRINFO) && defined(HAVE_GETADDRINFO)
 	struct addrinfo *res = NULL;
 	int result;
 #else
-        struct hostent *he;
+	struct hostent *he;
 #endif
 
-        ddebug("get_address()");
-        if (have_ipv6 && inet_pton(AF_INET6, host, &in6) == 1)
-                isc_sockaddr_fromin6(sockaddr, &in6, port);
-        else if (inet_pton(AF_INET, host, &in4) == 1)
-                isc_sockaddr_fromin(sockaddr, &in4, port);
-        else {
+	ddebug("get_address()");
+	if (have_ipv6 && inet_pton(AF_INET6, host, &in6) == 1)
+		isc_sockaddr_fromin6(sockaddr, &in6, port);
+	else if (inet_pton(AF_INET, host, &in4) == 1)
+		isc_sockaddr_fromin(sockaddr, &in4, port);
+	else {
 #if defined(HAVE_ADDRINFO) && defined(HAVE_GETADDRINFO)
 		result = getaddrinfo(host, NULL, NULL, &res);
 		if (result != 0) {
@@ -509,21 +517,22 @@ get_address(char *host, in_port_t port, isc_sockaddr_t *sockaddr) {
 		isc_sockaddr_setport(sockaddr, port);
 		freeaddrinfo(res);
 #else
-                he = gethostbyname(host);
-                if (he == NULL)
-                     fatal("Couldn't look up your server host %s.  errno=%d",
-                              host, h_errno);
-                INSIST(he->h_addrtype == AF_INET);
-                isc_sockaddr_fromin(sockaddr,
-                                    (struct in_addr *)(he->h_addr_list[0]),
-                                    port);
+		he = gethostbyname(host);
+		if (he == NULL)
+		     fatal("Couldn't look up your server host %s.  errno=%d",
+			      host, h_errno);
+		INSIST(he->h_addrtype == AF_INET);
+		isc_sockaddr_fromin(sockaddr,
+				    (struct in_addr *)(he->h_addr_list[0]),
+				    port);
 #endif
-        }
+	}
 }
 
 static void
 parse_args(int argc, char **argv) {
 	int ch;
+	isc_result_t result;
 
 	debug("parse_args");
 	while ((ch = isc_commandline_parse(argc, argv, "dDMy:vk:")) != -1) {
@@ -538,7 +547,9 @@ parse_args(int argc, char **argv) {
 		case 'M': /* was -dm */
 			debugging = ISC_TRUE;
 			ddebugging = ISC_TRUE;
-			isc_mem_debugging = ISC_TRUE;
+			memdebugging = ISC_TRUE;
+			isc_mem_debugging = ISC_MEM_DEBUGTRACE |
+					    ISC_MEM_DEBUGRECORD;
 			break;
 		case 'y':
 			keystr = isc_commandline_argument;
@@ -562,6 +573,22 @@ parse_args(int argc, char **argv) {
 			argv[0]);
 		exit(1);
 	}
+
+	if (argv[isc_commandline_index] != NULL) {
+		if (strcmp(argv[isc_commandline_index], "-") == 0) {
+			input = stdin;
+		} else {
+			result = isc_stdio_open(argv[isc_commandline_index],
+						"r", &input);
+			if (result != ISC_R_SUCCESS) {
+				fprintf(stderr, "isc_stdio_open(%s): %s\n",
+					argv[isc_commandline_index],
+					isc_result_totext(result));
+				exit(1);
+			}
+		}
+		interactive = ISC_FALSE;
+	}
 }
 
 static isc_uint16_t
@@ -606,7 +633,7 @@ parse_name(char **cmdlinep, dns_message_t *msg, dns_name_t **namep) {
 static isc_uint16_t
 parse_rdata(char **cmdlinep, dns_rdataclass_t rdataclass,
 	    dns_rdatatype_t rdatatype, dns_message_t *msg,
-	    dns_rdata_t **rdatap)
+	    dns_rdata_t *rdata)
 {
 	char *cmdline = *cmdlinep;
 	isc_buffer_t source, *buf = NULL;
@@ -614,33 +641,46 @@ parse_rdata(char **cmdlinep, dns_rdataclass_t rdataclass,
 	dns_rdatacallbacks_t callbacks;
 	isc_result_t result;
 	dns_name_t *rn;
+	int bufsz = INITDATA;
 
 	while (*cmdline != 0 && isspace((unsigned char)*cmdline))
 		cmdline++;
 
 	if (*cmdline != 0) {
-		result = isc_lex_create(mctx, WORDLEN, &lex);
-		check_result(result, "isc_lex_create");	
-
-		isc_buffer_init(&source, cmdline, strlen(cmdline));
-		isc_buffer_add(&source, strlen(cmdline));
-		result = isc_lex_openbuffer(lex, &source);
-		check_result(result, "isc_lex_openbuffer");
-
-		result = isc_buffer_allocate(mctx, &buf, MXNAME);
-		check_result(result, "isc_buffer_allocate");
-		dns_rdatacallbacks_init_stdio(&callbacks);
+		dns_rdatacallbacks_init(&callbacks);
 		if (userzone != NULL)
 			rn = userzone;
 		else
 			rn = origin;
-		result = dns_rdata_fromtext(*rdatap, rdataclass, rdatatype,
-					    lex, rn, ISC_FALSE, buf,
-					    &callbacks);
+		do {
+			result = isc_lex_create(mctx, strlen(cmdline), &lex);
+			check_result(result, "isc_lex_create");
+			isc_buffer_init(&source, cmdline, strlen(cmdline));
+			isc_buffer_add(&source, strlen(cmdline));
+			result = isc_lex_openbuffer(lex, &source);
+			check_result(result, "isc_lex_openbuffer");
+			if (buf != NULL)
+				isc_buffer_free(&buf);
+			if (bufsz > MAXDATA) {
+				fprintf(stderr, "couldn't allocate enough "
+					"space for the rdata\n");
+				exit(1);
+			}
+			result = isc_buffer_allocate(mctx, &buf, bufsz);
+			check_result(result, "isc_buffer_allocate");
+			result = dns_rdata_fromtext(rdata, rdataclass,
+						    rdatatype,
+						    lex, rn, ISC_FALSE, mctx,
+						    buf, &callbacks);
+			bufsz *= 2;
+			isc_lex_destroy(&lex);
+		} while (result == ISC_R_NOSPACE);
+		check_result(result, "dns_rdata_fromtext");
 		dns_message_takebuffer(msg, &buf);
-		isc_lex_destroy(&lex);
 		if (result != ISC_R_SUCCESS)
 			return (STATUS_MORE);
+	} else {
+		rdata->flags = DNS_RDATA_UPDATE;
 	}
 	*cmdlinep = cmdline;
 	return (STATUS_MORE);
@@ -711,7 +751,7 @@ make_prereq(char *cmdline, isc_boolean_t ispositive, isc_boolean_t isrrset) {
 
 	if (isrrset && ispositive) {
 		retval = parse_rdata(&cmdline, rdataclass, rdatatype,
-				     updatemsg, &rdata);
+				     updatemsg, rdata);
 		if (retval != STATUS_MORE)
 			return (retval);
 	}
@@ -736,7 +776,7 @@ make_prereq(char *cmdline, isc_boolean_t ispositive, isc_boolean_t isrrset) {
 	ISC_LIST_INIT(rdatalist->rdata);
 	ISC_LIST_APPEND(rdatalist->rdata, rdata, link);
 	dns_rdataset_init(rdataset);
-	dns_rdatalist_tordataset(rdatalist, rdataset);		
+	dns_rdatalist_tordataset(rdatalist, rdataset);
 	ISC_LIST_INIT(name->list);
 	ISC_LIST_APPEND(name->list, rdataset, link);
 	dns_message_addname(updatemsg, name, DNS_SECTION_PREREQUISITE);
@@ -812,6 +852,45 @@ evaluate_server(char *cmdline) {
 	return (STATUS_MORE);
 }
 
+static isc_uint16_t
+evaluate_local(char *cmdline) {
+	char *word, *local;
+	long port;
+
+	word = nsu_strsep(&cmdline, " \t\r\n");
+	if (*word == 0) {
+		fprintf(stderr, "failed to read server name\n");
+		return (STATUS_SYNTAX);
+	}
+	local = word;
+
+	word = nsu_strsep(&cmdline, " \t\r\n");
+	if (*word == 0)
+		port = 0;
+	else {
+		char *endp;
+		port = strtol(word, &endp, 10);
+		if (*endp != 0) {
+			fprintf(stderr, "port '%s' is not numeric\n", word);
+			return (STATUS_SYNTAX);
+		} else if (port < 1 || port > 65535) {
+			fprintf(stderr, "port '%s' is out of range "
+				"(1 to 65535)\n", word);
+			return (STATUS_SYNTAX);
+		}
+	}
+
+	if (localaddr == NULL) {
+		localaddr = isc_mem_get(mctx, sizeof(isc_sockaddr_t));
+		if (localaddr == NULL)
+			fatal("out of memory");
+	}
+
+	get_address(local, (in_port_t)port, localaddr);
+
+	return (STATUS_MORE);
+}
+
 static isc_uint16_t
 evaluate_zone(char *cmdline) {
 	char *word;
@@ -883,7 +962,7 @@ update_addordelete(char *cmdline, isc_boolean_t isdelete) {
 		if (*endp != '\0') {
 			fprintf(stderr, "ttl '%s' is not numeric\n", word);
 			goto failure;
-		} else if (ttl < 1 || ttl > TTL_MAX || errno == ERANGE) {
+		} else if (ttl < 0 || ttl > TTL_MAX || errno == ERANGE) {
 			/*
 			 * The errno test is needed to catch when strtol()
 			 * overflows on a platform where sizeof(int) ==
@@ -891,7 +970,7 @@ update_addordelete(char *cmdline, isc_boolean_t isdelete) {
 			 * which will be equal to TTL_MAX.
 			 */
 			fprintf(stderr, "ttl '%s' is out of range "
-				"(1 to %d)\n", word, TTL_MAX);
+				"(0 to %d)\n", word, TTL_MAX);
 			goto failure;
 		}
 	} else
@@ -905,6 +984,7 @@ update_addordelete(char *cmdline, isc_boolean_t isdelete) {
 		if (isdelete) {
 			rdataclass = dns_rdataclass_any;
 			rdatatype = dns_rdatatype_any;
+			rdata->flags = DNS_RDATA_UPDATE;
 			goto doneparsing;
 		} else {
 			fprintf(stderr, "failed to read class or type\n");
@@ -923,6 +1003,7 @@ update_addordelete(char *cmdline, isc_boolean_t isdelete) {
 			if (isdelete) {
 				rdataclass = dns_rdataclass_any;
 				rdatatype = dns_rdatatype_any;
+				rdata->flags = DNS_RDATA_UPDATE;
 				goto doneparsing;
 			} else {
 				fprintf(stderr, "failed to read type\n");
@@ -940,17 +1021,17 @@ update_addordelete(char *cmdline, isc_boolean_t isdelete) {
 	}
 
 	retval = parse_rdata(&cmdline, rdataclass, rdatatype, updatemsg,
-			     &rdata);
+			     rdata);
 	if (retval != STATUS_MORE)
 		goto failure;
 
 	if (isdelete) {
-		if (rdata->length == 0)
+		if ((rdata->flags & DNS_RDATA_UPDATE) != 0)
 			rdataclass = dns_rdataclass_any;
 		else
 			rdataclass = dns_rdataclass_none;
 	} else {
-		if (rdata->length == 0) {
+		if ((rdata->flags & DNS_RDATA_UPDATE) != 0) {
 			fprintf(stderr, "failed to read rdata\n");
 			goto failure;
 		}
@@ -1007,23 +1088,37 @@ evaluate_update(char *cmdline) {
 }
 
 static void
-show_message(void) {
+show_message(dns_message_t *msg) {
 	isc_result_t result;
-	char store[MSGTEXT];
-	isc_buffer_t buf;
+	isc_buffer_t *buf = NULL;
+	int bufsz;
 
 	ddebug("show_message()");
-	isc_buffer_init(&buf, store, MSGTEXT);
-	result = dns_message_totext(updatemsg, 0, &buf);
+	bufsz = INITTEXT;
+	do { 
+		if (bufsz > MAXTEXT) {
+			fprintf(stderr, "couldn't allocate large enough "
+				"buffer to display message\n");
+			exit(1);
+		}
+		if (buf != NULL)
+			isc_buffer_free(&buf);
+		result = isc_buffer_allocate(mctx, &buf, bufsz);
+		check_result(result, "isc_buffer_allocate");
+		result = dns_message_totext(msg, 0, buf);
+		bufsz *= 2;
+	} while (result == ISC_R_NOSPACE);
 	if (result != ISC_R_SUCCESS) {
-		fprintf(stderr, "Failed to concert message to text format.\n");
+		fprintf(stderr, "Failed to convert message to text format.\n");
+		isc_buffer_free(&buf);
 		return;
 	}
 	printf("Outgoing update query:\n%.*s",
-	       (int)isc_buffer_usedlength(&buf),
-	       (char*)isc_buffer_base(&buf));
+	       (int)isc_buffer_usedlength(buf),
+	       (char*)isc_buffer_base(buf));
+	isc_buffer_free(&buf);
 }
-	
+
 
 static isc_uint16_t
 get_next_command(void) {
@@ -1032,12 +1127,14 @@ get_next_command(void) {
 	char *word;
 
 	ddebug("get_next_command()");
-	fprintf(stdout, "> ");
-	fgets (cmdlinebuf, MAXCMD, stdin);
-	cmdline = cmdlinebuf;
+	if (interactive)
+		fprintf(stdout, "> ");
+	cmdline = fgets(cmdlinebuf, MAXCMD, input);
+	if (cmdline == NULL)
+		return (STATUS_QUIT);
 	word = nsu_strsep(&cmdline, " \t\r\n");
 
-	if (feof(stdin))
+	if (feof(input))
 		return (STATUS_QUIT);
 	if (*word == 0)
 		return (STATUS_SEND);
@@ -1049,12 +1146,14 @@ get_next_command(void) {
 		return (evaluate_update(cmdline));
 	if (strcasecmp(word, "server") == 0)
 		return (evaluate_server(cmdline));
+	if (strcasecmp(word, "local") == 0)
+		return (evaluate_local(cmdline));
 	if (strcasecmp(word, "zone") == 0)
 		return (evaluate_zone(cmdline));
 	if (strcasecmp(word, "send") == 0)
 		return (STATUS_SEND);
 	if (strcasecmp(word, "show") == 0) {
-		show_message();
+		show_message(updatemsg);
 		return (STATUS_MORE);
 	}
 	fprintf(stderr, "incorrect section name: %s\n", word);
@@ -1075,55 +1174,72 @@ user_interaction(void) {
 }
 
 static void
-done_update(isc_boolean_t acquirelock) {
-	if (acquirelock)
-		LOCK(&lock);
-	busy = ISC_FALSE;
-	SIGNAL(&cond);
-	if (acquirelock)
-		UNLOCK(&lock);
+done_update(void) {
+	isc_event_t *event = global_event;
+	ddebug("done_update()");
+	isc_task_send(global_task, &event);
 }
 
 static void
 update_completed(isc_task_t *task, isc_event_t *event) {
 	dns_requestevent_t *reqev = NULL;
 	isc_result_t result;
-	isc_buffer_t buf;
 	dns_message_t *rcvmsg = NULL;
-	char bufstore[MSGTEXT];
-	
+
 	UNUSED(task);
 
-	ddebug("updated_completed()");
+	ddebug("update_completed()");
 	REQUIRE(event->ev_type == DNS_EVENT_REQUESTDONE);
 	reqev = (dns_requestevent_t *)event;
 	if (reqev->result != ISC_R_SUCCESS) {
 		fprintf(stderr, "; Communication with server failed: %s\n",
 			isc_result_totext(reqev->result));
+		seenerror = ISC_TRUE;
 		goto done;
 	}
 
 	result = dns_message_create(mctx, DNS_MESSAGE_INTENTPARSE, &rcvmsg);
 	check_result(result, "dns_message_create");
-	result = dns_request_getresponse(reqev->request, rcvmsg, ISC_TRUE);
+	result = dns_request_getresponse(reqev->request, rcvmsg,
+					 DNS_MESSAGEPARSE_PRESERVEORDER);
 	check_result(result, "dns_request_getresponse");
+	if (rcvmsg->rcode != dns_rcode_noerror)
+		seenerror = ISC_TRUE;
 	if (debugging) {
-		isc_buffer_init(&buf, bufstore, MSGTEXT);
-		result = dns_message_totext(rcvmsg, 0, &buf);
+		isc_buffer_t *buf = NULL;
+		int bufsz;
+
+		bufsz = INITTEXT;
+		do { 
+			if (bufsz > MAXTEXT) {
+				fprintf(stderr, "couldn't allocate large "
+					"enough buffer to display message\n");
+				exit(1);
+			}
+			if (buf != NULL)
+				isc_buffer_free(&buf);
+			result = isc_buffer_allocate(mctx, &buf, bufsz);
+			check_result(result, "isc_buffer_allocate");
+			result = dns_message_totext(rcvmsg, 0, buf);
+			bufsz *= 2;
+		} while (result == ISC_R_NOSPACE);
 		check_result(result, "dns_message_totext");
 		fprintf(stderr, "\nReply from update query:\n%.*s\n",
-			(int)isc_buffer_usedlength(&buf),
-			(char*)isc_buffer_base(&buf));
+			(int)isc_buffer_usedlength(buf),
+			(char*)isc_buffer_base(buf));
+		isc_buffer_free(&buf);
 	}
 	dns_message_destroy(&rcvmsg);
  done:
 	dns_request_destroy(&reqev->request);
 	isc_event_free(&event);
-	done_update(ISC_TRUE);
+	done_update();
 }
 
 static void
-send_update(dns_name_t *zonename, isc_sockaddr_t *master) {
+send_update(dns_name_t *zonename, isc_sockaddr_t *master,
+	    isc_sockaddr_t *srcaddr)
+{
 	isc_result_t result;
 	dns_request_t *request = NULL;
 	dns_name_t *name = NULL;
@@ -1146,11 +1262,11 @@ send_update(dns_name_t *zonename, isc_sockaddr_t *master) {
 
 	if (usevc)
 		options |= DNS_REQUESTOPT_TCP;
-	result = dns_request_create(requestmgr, updatemsg, master,
-				    options, key,
-				    FIND_TIMEOUT, global_task,
-				    update_completed, NULL, &request);
-	check_result(result, "dns_request_create");
+	result = dns_request_createvia(requestmgr, updatemsg, srcaddr,
+				       master, options, key,
+				       FIND_TIMEOUT, global_task,
+				       update_completed, NULL, &request);
+	check_result(result, "dns_request_createvia");
 }
 
 static void
@@ -1163,7 +1279,7 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 	dns_name_t *name = NULL;
 	dns_rdataset_t *soaset = NULL;
 	dns_rdata_soa_t soa;
-	dns_rdata_t soarr;
+	dns_rdata_t soarr = DNS_RDATA_INIT;
 	int pass = 0;
 	dns_name_t master;
 	isc_sockaddr_t *serveraddr, tempaddr;
@@ -1171,6 +1287,7 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 	nsu_requestinfo_t *reqinfo;
 	dns_message_t *soaquery = NULL;
 	isc_sockaddr_t *addr;
+	isc_boolean_t seencname = ISC_FALSE;
 
 	UNUSED(task);
 
@@ -1188,40 +1305,51 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 
 	if (eresult != ISC_R_SUCCESS) {
 		char addrbuf[ISC_SOCKADDR_FORMATSIZE];
-	
+
 		isc_sockaddr_format(addr, addrbuf, sizeof(addrbuf));
 		fprintf(stderr, "; Communication with %s failed: %s\n",
 		       addrbuf, isc_result_totext(eresult));
 		if (userserver != NULL)
 			fatal("Couldn't talk to specified nameserver.");
-		else if (ns_inuse++ >= lwconf->nsnext)
+		else if (++ns_inuse >= lwconf->nsnext)
 			fatal("Couldn't talk to any default nameserver.");
 		ddebug("Destroying request [%lx]", request);
 		dns_request_destroy(&request);
 		dns_message_renderreset(soaquery);
-		sendrequest(&servers[ns_inuse], soaquery, &request);
+		sendrequest(localaddr, &servers[ns_inuse], soaquery, &request);
 		isc_mem_put(mctx, reqinfo, sizeof(nsu_requestinfo_t));
 		return;
 	}
-	dns_message_destroy(&soaquery);
 	isc_mem_put(mctx, reqinfo, sizeof(nsu_requestinfo_t));
 
 	ddebug("About to create rcvmsg");
 	result = dns_message_create(mctx, DNS_MESSAGE_INTENTPARSE, &rcvmsg);
 	check_result(result, "dns_message_create");
-	result = dns_request_getresponse(request, rcvmsg, ISC_TRUE);
+	result = dns_request_getresponse(request, rcvmsg,
+					 DNS_MESSAGEPARSE_PRESERVEORDER);
 	check_result(result, "dns_request_getresponse");
 	section = DNS_SECTION_ANSWER;
 	if (debugging) {
-		isc_buffer_t buf;
-		char bufstore[MSGTEXT];
-
-		isc_buffer_init(&buf, bufstore, MSGTEXT);
-		result = dns_message_totext(rcvmsg, 0, &buf);
+		isc_buffer_t *buf = NULL;
+		int bufsz;
+		bufsz = INITTEXT;
+		do {
+			if (buf != NULL)
+				isc_buffer_free(&buf);
+			if (bufsz > MAXTEXT) {
+				fprintf(stderr, "couldn't allocate enough "
+					 "space for debugging message\n");
+				exit(1);
+			}
+			result = isc_buffer_allocate(mctx, &buf, bufsz);
+			check_result(result, "isc_buffer_allocate");
+			result = dns_message_totext(rcvmsg, 0, buf);
+		} while (result == ISC_R_NOSPACE);
 		check_result(result, "dns_message_totext");
 		fprintf(stderr, "Reply from SOA query:\n%.*s\n",
-			(int)isc_buffer_usedlength(&buf),
-			(char*)isc_buffer_base(&buf));
+			(int)isc_buffer_usedlength(buf),
+			(char*)isc_buffer_base(buf));
+		isc_buffer_free(&buf);
 	}
 
 	if (rcvmsg->rcode != dns_rcode_noerror &&
@@ -1250,16 +1378,54 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 					      &soaset);
 		if (result == ISC_R_SUCCESS)
 			break;
+		if (section == DNS_SECTION_ANSWER) {
+			dns_rdataset_t *tset = NULL;
+			if (dns_message_findtype(name, dns_rdatatype_cname, 0,
+						 &tset) == ISC_R_SUCCESS
+			    ||
+			    dns_message_findtype(name, dns_rdatatype_dname, 0,
+						 &tset) == ISC_R_SUCCESS
+			    )
+			{
+				seencname = ISC_TRUE;
+				break;
+			}
+		}
+				
 		result = dns_message_nextname(rcvmsg, section);
 	}
 
-	if (soaset == NULL) {
+	if (soaset == NULL && !seencname) {
 		pass++;
 		goto lookforsoa;
 	}
 
+	if (seencname) {
+		dns_name_t tname;
+		unsigned int nlabels;
+
+		result = dns_message_firstname(soaquery, DNS_SECTION_QUESTION);
+		INSIST(result == ISC_R_SUCCESS);
+		name = NULL;
+		dns_message_currentname(soaquery, DNS_SECTION_QUESTION, &name);
+		nlabels = dns_name_countlabels(name);
+		if (nlabels == 1)
+			fatal("could not find enclosing zone");
+		dns_name_init(&tname, NULL);
+		dns_name_getlabelsequence(name, 1, nlabels - 1, &tname);
+		dns_name_clone(&tname, name);
+		dns_request_destroy(&request);
+		dns_message_renderreset(soaquery);
+		if (userserver != NULL)
+			sendrequest(localaddr, userserver, soaquery, &request);
+		else
+			sendrequest(localaddr, &servers[ns_inuse], soaquery,
+				    &request);
+		goto out;
+	}
+
 	if (debugging) {
-		char namestr[MAXPNAME];
+		char namestr[DNS_NAME_FORMATSIZE];
 		dns_name_format(name, namestr, sizeof(namestr));
 		fprintf(stderr, "Found zone name: %s\n", namestr);
 	}
@@ -1281,7 +1447,7 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 		zonename = name;
 
 	if (debugging) {
-		char namestr[MAXPNAME];
+		char namestr[DNS_NAME_FORMATSIZE];
 		dns_name_format(&master, namestr, sizeof(namestr));
 		fprintf(stderr, "The master is: %s\n", namestr);
 	}
@@ -1289,7 +1455,7 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 	if (userserver != NULL)
 		serveraddr = userserver;
 	else {
-		char serverstr[MXNAME];
+		char serverstr[DNS_NAME_MAXTEXT+1];
 		isc_buffer_t buf;
 
 		isc_buffer_init(&buf, serverstr, sizeof(serverstr));
@@ -1300,17 +1466,20 @@ recvsoa(isc_task_t *task, isc_event_t *event) {
 		serveraddr = &tempaddr;
 	}
 
-	send_update(zonename, serveraddr);
+	send_update(zonename, serveraddr, localaddr);
 
+	dns_message_destroy(&soaquery);
+	dns_request_destroy(&request);
+
+ out:
 	dns_rdata_freestruct(&soa);
 	dns_message_destroy(&rcvmsg);
-	dns_request_destroy(&request);
 	ddebug("Out of recvsoa");
 }
 
 static void
-sendrequest(isc_sockaddr_t *address, dns_message_t *msg,
-	    dns_request_t **request)
+sendrequest(isc_sockaddr_t *srcaddr, isc_sockaddr_t *destaddr,
+	    dns_message_t *msg, dns_request_t **request)
 {
 	isc_result_t result;
 	nsu_requestinfo_t *reqinfo;
@@ -1319,11 +1488,11 @@ sendrequest(isc_sockaddr_t *address, dns_message_t *msg,
 	if (reqinfo == NULL)
 		fatal("out of memory");
 	reqinfo->msg = msg;
-	reqinfo->addr = address;
-	result = dns_request_create(requestmgr, msg, address,
-				    0, NULL, FIND_TIMEOUT, global_task,
-				    recvsoa, reqinfo, request);
-	check_result(result, "dns_request_create");
+	reqinfo->addr = destaddr;
+	result = dns_request_createvia(requestmgr, msg, srcaddr, destaddr,
+				       0, NULL, FIND_TIMEOUT, global_task,
+				       recvsoa, reqinfo, request);
+	check_result(result, "dns_request_createvia");
 }
 
 static void
@@ -1338,13 +1507,13 @@ start_update(void) {
 	ddebug("start_update()");
 
 	if (userzone != NULL && userserver != NULL) {
-		send_update(userzone, userserver);
+		send_update(userzone, userserver, localaddr);
 		return;
 	}
 
 	result = dns_message_firstname(updatemsg, DNS_SECTION_UPDATE);
 	if (result != ISC_R_SUCCESS) {
-		done_update(ISC_FALSE);
+		done_update();
 		return;
 	}
 
@@ -1373,10 +1542,10 @@ start_update(void) {
 	dns_message_addname(soaquery, name, DNS_SECTION_QUESTION);
 
 	if (userserver != NULL)
-		sendrequest(userserver, soaquery, &request);
+		sendrequest(localaddr, userserver, soaquery, &request);
 	else {
 		ns_inuse = 0;
-		sendrequest(&servers[ns_inuse], soaquery, &request);
+		sendrequest(localaddr, &servers[ns_inuse], soaquery, &request);
 	}
 }
 
@@ -1387,17 +1556,14 @@ cleanup(void) {
 	if (userserver != NULL)
 		isc_mem_put(mctx, userserver, sizeof(isc_sockaddr_t));
 
+	if (localaddr != NULL)
+		isc_mem_put(mctx, localaddr, sizeof(isc_sockaddr_t));
+
 	if (key != NULL) {
 		debug("Freeing key");
-		dns_tsigkey_setdeleted(key);
 		dns_tsigkey_detach(&key);
 	}
 
-	if (keyring != NULL) {
-		debug("Freeing keyring %lx", keyring);
-		dns_tsigkeyring_destroy(&keyring);
-	}
-
 	if (updatemsg != NULL)
 		dns_message_destroy(&updatemsg);
 
@@ -1416,25 +1582,26 @@ cleanup(void) {
 	lwres_context_destroy(&lwctx);
 
 	isc_mem_put(mctx, servers, ns_total * sizeof(isc_sockaddr_t));
-		
+
 	ddebug("Shutting down request manager");
 	dns_requestmgr_shutdown(requestmgr);
 	dns_requestmgr_detach(&requestmgr);
 
 	ddebug("Freeing the dispatchers");
-	dns_dispatch_detach(&dispatchv4);
+	if (have_ipv4)
+		dns_dispatch_detach(&dispatchv4);
 	if (have_ipv6)
 		dns_dispatch_detach(&dispatchv6);
 
 	ddebug("Shutting down dispatch manager");
 	dns_dispatchmgr_destroy(&dispatchmgr);
 
-	ddebug("Ending task");
-	isc_task_detach(&global_task);
-
 	ddebug("Shutting down task manager");
 	isc_taskmgr_destroy(&taskmgr);
 
+	ddebug("Destroying event");
+	isc_event_free(&global_event);
+
 	ddebug("Shutting down socket manager");
 	isc_socketmgr_destroy(&socketmgr);
 
@@ -1442,38 +1609,59 @@ cleanup(void) {
 	isc_timermgr_destroy(&timermgr);
 
 	ddebug("Destroying memory context");
-	if (isc_mem_debugging)
+	if (memdebugging)
 		isc_mem_stats(mctx, stderr);
 	isc_mem_destroy(&mctx);
 }
 
+static void
+getinput(isc_task_t *task, isc_event_t *event) {
+	isc_boolean_t more;
+
+	UNUSED(task);
+
+	if (global_event == NULL)
+		global_event = event;
+
+	if (shuttingdown) {
+		isc_app_shutdown();
+		return;
+	}
+	reset_system();
+	isc_app_block();
+	more = user_interaction();
+	isc_app_unblock();
+	if (!more) {
+		isc_app_shutdown();
+		return;
+	}
+	start_update();
+	return;
+}
+
 int
 main(int argc, char **argv) {
-        isc_result_t result;
+	isc_result_t result;
 
-        parse_args(argc, argv);
+	input = stdin;
 
-        setup_system();
-        result = isc_mutex_init(&lock);
-        check_result(result, "isc_mutex_init");
-        result = isc_condition_init(&cond);
-        check_result(result, "isc_condition_init");
-	LOCK(&lock);
+	isc_app_start();
 
-        while (ISC_TRUE) {
-		reset_system();
-                if (!user_interaction())
-			break;
-		busy = ISC_TRUE;
-		start_update();
-		while (busy)
-			WAIT(&cond, &lock);
-        }
+	parse_args(argc, argv);
 
-        fprintf(stdout, "\n");
-        isc_mutex_destroy(&lock);
-        isc_condition_destroy(&cond);
-        cleanup();
+	setup_system();
 
-        return (0);
+	result = isc_app_onrun(mctx, global_task, getinput, NULL);
+	check_result(result, "isc_app_onrun");
+
+	(void)isc_app_run();
+
+	cleanup();
+
+	isc_app_finish();
+
+	if (seenerror)
+		return (2);
+	else
+		return (0);
 }
