@@ -1,23 +1,22 @@
 /*
- * Copyright (C) 1999, 2000  Internet Software Consortium.
- * 
+ * Copyright (C) 1999-2001  Internet Software Consortium.
+ *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
- * 
- * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
- * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
- * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
- * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
- * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
- * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
- * SOFTWARE.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM
+ * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
+ * INTERNET SOFTWARE CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT,
+ * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
+ * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
+ * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
+ * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 /*
- * $Id: tkey.c,v 1.47 2000/06/23 01:51:35 bwelling Exp $
- * Principal Author: Brian Wellington
+ * $Id: tkey.c,v 1.57.2.4 2001/01/11 20:34:09 gson Exp $
  */
 
 #include <config.h>
@@ -42,6 +41,9 @@
 #include <dns/tkey.h>
 #include <dns/tsig.h>
 
+#include <dst/dst.h>
+#include <dst/gssapi.h>
+
 #define TKEY_RANDOM_AMOUNT 16
 
 #define RETERR(x) do { \
@@ -50,12 +52,6 @@
 		goto failure; \
 	} while (0)
 
-#define TKEYTRACE(m)	isc_log_write(dns_lctx, \
-				      DNS_LOGCATEGORY_RESOLVER, \
-				      DNS_LOGMODULE_RESOLVER, \
-				      ISC_LOG_DEBUG(3), \
-				      "tkey: %s", (m))
-
 static void
 tkey_log(const char *fmt, ...) {
 	va_list ap;
@@ -84,6 +80,7 @@ dns_tkeyctx_create(isc_mem_t *mctx, isc_entropy_t *ectx, dns_tkeyctx_t **tctxp)
 	isc_entropy_attach(ectx, &tctx->ectx);
 	tctx->dhkey = NULL;
 	tctx->domain = NULL;
+	tctx->gsscred = NULL;
 
 	*tctxp = tctx;
 	return (ISC_R_SUCCESS);
@@ -102,7 +99,8 @@ dns_tkeyctx_destroy(dns_tkeyctx_t **tctxp) {
 	if (tctx->dhkey != NULL)
 		dst_key_free(&tctx->dhkey);
 	if (tctx->domain != NULL) {
-		dns_name_free(tctx->domain, mctx);
+		if (dns_name_dynamic(tctx->domain))
+			dns_name_free(tctx->domain, mctx);
 		isc_mem_put(mctx, tctx->domain, sizeof(dns_name_t));
 	}
 	isc_entropy_detach(&tctx->ectx);
@@ -121,7 +119,7 @@ add_rdata_to_list(dns_message_t *msg, dns_name_t *name, dns_rdata_t *rdata,
 	dns_name_t *newname = NULL;
 	dns_rdatalist_t *newlist = NULL;
 	dns_rdataset_t *newset = NULL;
-	isc_buffer_t *tmprdatabuf = NULL, *tmpnamebuf = NULL;
+	isc_buffer_t *tmprdatabuf = NULL;
 
 	RETERR(dns_message_gettemprdata(msg, &newrdata));
 
@@ -132,14 +130,9 @@ add_rdata_to_list(dns_message_t *msg, dns_name_t *name, dns_rdata_t *rdata,
 	dns_rdata_fromregion(newrdata, rdata->rdclass, rdata->type, &newr);
 	dns_message_takebuffer(msg, &tmprdatabuf);
 
-	dns_name_toregion(name, &r);
 	RETERR(dns_message_gettempname(msg, &newname));
 	dns_name_init(newname, NULL);
-	RETERR(isc_buffer_allocate(msg->mctx, &tmpnamebuf, r.length));
-	isc_buffer_availableregion(tmpnamebuf, &newr);
-	memcpy(newr.base, r.base, r.length);
-	dns_name_fromregion(newname, &newr);
-	dns_message_takebuffer(msg, &tmpnamebuf);
+	RETERR(dns_name_dup(name, msg->mctx, newname));
 
 	RETERR(dns_message_gettemprdatalist(msg, &newlist));
 	newlist->rdclass = newrdata->rdclass;
@@ -161,19 +154,39 @@ add_rdata_to_list(dns_message_t *msg, dns_name_t *name, dns_rdata_t *rdata,
 	return (ISC_R_SUCCESS);
 
  failure:
-	if (newrdata != NULL)
+	if (newrdata != NULL) {
+		if (ISC_LINK_LINKED(newrdata, link))
+			ISC_LIST_UNLINK(newlist->rdata, newrdata, link);
 		dns_message_puttemprdata(msg, &newrdata);
+	}
 	if (newname != NULL)
 		dns_message_puttempname(msg, &newname);
-	if (newlist != NULL)
-		dns_message_puttemprdatalist(msg, &newlist);
 	if (newset != NULL) {
 		dns_rdataset_disassociate(newset);
 		dns_message_puttemprdataset(msg, &newset);
 	}
+	if (newlist != NULL)
+		dns_message_puttemprdatalist(msg, &newlist);
 	return (result);
 }
 
+static void
+free_namelist(dns_message_t *msg, dns_namelist_t *namelist) {
+	dns_name_t *name;
+	dns_rdataset_t *set;
+
+	while (!ISC_LIST_EMPTY(*namelist)) {
+		name = ISC_LIST_HEAD(*namelist);
+		ISC_LIST_UNLINK(*namelist, name, link);
+		while (!ISC_LIST_EMPTY(name->list)) {
+			set = ISC_LIST_HEAD(name->list);
+			ISC_LIST_UNLINK(name->list, set, link);
+			dns_message_puttemprdataset(msg, &set);
+		}
+		dns_message_puttempname(msg, &name);
+	}
+}
+
 static isc_result_t
 compute_secret(isc_buffer_t *shared, isc_region_t *queryrandomness,
 	       isc_region_t *serverrandomness, isc_buffer_t *secret)
@@ -234,7 +247,7 @@ process_dhtkey(dns_message_t *msg, dns_name_t *signer, dns_name_t *name,
 	isc_result_t result = ISC_R_SUCCESS;
 	dns_name_t *keyname, ourname;
 	dns_rdataset_t *keyset;
-	dns_rdata_t keyrdata, ourkeyrdata;
+	dns_rdata_t keyrdata = DNS_RDATA_INIT, ourkeyrdata = DNS_RDATA_INIT;
 	isc_boolean_t found_key = ISC_FALSE, found_incompatible = ISC_FALSE;
 	dst_key_t *pubkey = NULL;
 	isc_buffer_t ourkeybuf, *shared = NULL;
@@ -246,6 +259,19 @@ process_dhtkey(dns_message_t *msg, dns_name_t *signer, dns_name_t *name,
 	unsigned char *randomdata = NULL, secretdata[256];
 	isc_stdtime_t now;
 
+	if (tctx->dhkey == NULL) {
+		tkey_log("process_dhtkey: tkey-dhkey not defined");
+		tkeyout->error = dns_tsigerror_badalg;
+		return (DNS_R_REFUSED);
+	}
+
+	if (!dns_name_equal(&tkeyin->algorithm, DNS_TSIG_HMACMD5_NAME)) {
+		tkey_log("process_dhtkey: algorithms other than "
+			 "hmac-md5 are not supported");
+		tkeyout->error = dns_tsigerror_badalg;
+		return (ISC_R_SUCCESS);
+	}
+
 	/*
 	 * Look for a DH KEY record that will work with ours.
 	 */
@@ -410,6 +436,78 @@ process_dhtkey(dns_message_t *msg, dns_name_t *signer, dns_name_t *name,
 	return (result);
 }
 
+static isc_result_t
+process_gsstkey(dns_message_t *msg, dns_name_t *signer, dns_name_t *name,
+		dns_rdata_tkey_t *tkeyin, dns_tkeyctx_t *tctx,
+		dns_rdata_tkey_t *tkeyout,
+		dns_tsig_keyring_t *ring, dns_namelist_t *namelist)
+{
+	isc_result_t result = ISC_R_SUCCESS;
+	dst_key_t *dstkey = NULL;
+	void *gssctx = NULL;
+	isc_stdtime_t now;
+	isc_region_t intoken;
+	unsigned char array[1024];
+	isc_buffer_t outtoken;
+
+	UNUSED(namelist);
+
+	if (tctx->gsscred == NULL)
+		return (ISC_R_NOPERM);
+
+	if (!dns_name_equal(&tkeyin->algorithm, DNS_TSIG_GSSAPI_NAME) &&
+	    !dns_name_equal(&tkeyin->algorithm, DNS_TSIG_GSSAPIMS_NAME)) {
+		tkeyout->error = dns_tsigerror_badalg;
+		return (ISC_R_SUCCESS);
+	}
+
+	intoken.base = tkeyin->key;
+	intoken.length = tkeyin->keylen;
+
+	isc_buffer_init(&outtoken, array, sizeof(array));
+	RETERR(dst_gssapi_acceptctx(name, tctx->gsscred, &intoken,
+				    &outtoken, &gssctx));
+
+	dstkey = NULL;
+	RETERR(dst_key_fromgssapi(name, gssctx, msg->mctx, &dstkey));
+
+	result = dns_tsigkey_createfromkey(name, &tkeyin->algorithm,
+					   dstkey, ISC_TRUE, signer,
+					   tkeyin->inception, tkeyin->expire,
+					   msg->mctx, ring, NULL);
+	if (result != ISC_R_SUCCESS)
+		goto failure;
+
+	if (result == ISC_R_NOTFOUND) {
+		tkeyout->error = dns_tsigerror_badalg;
+		return (ISC_R_SUCCESS);
+	}
+	if (result != ISC_R_SUCCESS)
+		goto failure;
+
+	/* This key is good for a long time */
+	isc_stdtime_get(&now);
+	tkeyout->inception = tkeyin->inception;
+	tkeyout->expire = tkeyin->expire;
+
+	tkeyout->key = isc_mem_get(msg->mctx,
+				   isc_buffer_usedlength(&outtoken));
+	if (tkeyout->key == NULL) {
+		result = ISC_R_NOMEMORY;
+		goto failure;
+	}
+	tkeyout->keylen = isc_buffer_usedlength(&outtoken);
+	memcpy(tkeyout->key, isc_buffer_base(&outtoken), tkeyout->keylen);
+
+	return (ISC_R_SUCCESS);
+
+ failure:
+	if (dstkey != NULL)
+		dst_key_free(&dstkey);
+
+	return (result);
+}
+
 static isc_result_t
 process_deletetkey(dns_message_t *msg, dns_name_t *signer, dns_name_t *name,
 		   dns_rdata_tkey_t *tkeyin,
@@ -421,10 +519,9 @@ process_deletetkey(dns_message_t *msg, dns_name_t *signer, dns_name_t *name,
 	dns_tsigkey_t *tsigkey = NULL;
 	dns_name_t *identity;
 
-	/* Unused variables */
-	msg = msg;
-	tkeyout = tkeyout;
-	namelist = namelist;
+	UNUSED(msg);
+	UNUSED(tkeyout);
+	UNUSED(namelist);
 
 	result = dns_tsigkey_find(&tsigkey, name, &tkeyin->algorithm, ring);
 	if (result != ISC_R_SUCCESS)
@@ -456,9 +553,9 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 {
 	isc_result_t result = ISC_R_SUCCESS;
 	dns_rdata_tkey_t tkeyin, tkeyout;
-	dns_name_t *qname, *name, *keyname, tempkeyname, signer;
+	dns_name_t *qname, *name, *keyname, tempkeyname, *signer, tsigner;
 	dns_rdataset_t *tkeyset;
-	dns_rdata_t tkeyrdata, *rdata = NULL;
+	dns_rdata_t tkeyrdata = DNS_RDATA_INIT, *rdata = NULL;
 	isc_buffer_t *dynbuf = NULL;
 	dns_namelist_t namelist;
 	isc_boolean_t freealg = ISC_FALSE;
@@ -476,7 +573,8 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 	 * Interpret the question section.
 	 */
 	result = dns_message_firstname(msg, DNS_SECTION_QUESTION);
-	INSIST(result == ISC_R_SUCCESS);
+	if (result != ISC_R_SUCCESS)
+		return (DNS_R_FORMERR);
 
 	qname = NULL;
 	dns_message_currentname(msg, DNS_SECTION_QUESTION, &qname);
@@ -489,10 +587,19 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 	result = dns_message_findname(msg, DNS_SECTION_ADDITIONAL, qname,
 				      dns_rdatatype_tkey, 0, &name, &tkeyset);
 	if (result != ISC_R_SUCCESS) {
-		result = DNS_R_FORMERR;
-		tkey_log("dns_tkey_processquery: couldn't find a TKEY "
-			 "matching the question");
-		goto failure;
+		/*
+		 * Try the answer section, since that's where Win2000
+		 * puts it.
+		 */
+		if (dns_message_findname(msg, DNS_SECTION_ANSWER, qname,
+					 dns_rdatatype_tkey, 0, &name,
+					 &tkeyset) != ISC_R_SUCCESS)
+		{
+			result = DNS_R_FORMERR;
+			tkey_log("dns_tkey_processquery: couldn't find a TKEY "
+				 "matching the question");
+			goto failure;
+		}
 	}
 	result = dns_rdataset_first(tkeyset);
 	if (result != ISC_R_SUCCESS) {
@@ -501,7 +608,7 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 	}
 	dns_rdataset_current(tkeyset, &tkeyrdata);
 
-	RETERR(dns_rdata_tostruct(&tkeyrdata, &tkeyin, msg->mctx));
+	RETERR(dns_rdata_tostruct(&tkeyrdata, &tkeyin, NULL));
 
 	if (tkeyin.error != dns_rcode_noerror) {
 		result = DNS_R_FORMERR;
@@ -512,16 +619,20 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 	 * Before we go any farther, verify that the message was signed.
 	 * GSSAPI TKEY doesn't require a signature, the rest do.
 	 */
-	dns_name_init(&signer, NULL);
-	result = dns_message_signer(msg, &signer);
-	if (result != ISC_R_SUCCESS &&
-	    (tkeyin.mode != DNS_TKEYMODE_GSSAPI || result != ISC_R_NOTFOUND))
-	{
-		tkey_log("dns_tkey_processquery: query was not properly "
-			 "signed - rejecting");
-		result = DNS_R_FORMERR;
-		goto failure;
-	}
+	dns_name_init(&tsigner, NULL);
+	result = dns_message_signer(msg, &tsigner);
+	if (result != ISC_R_SUCCESS) {
+		if (tkeyin.mode == DNS_TKEYMODE_GSSAPI &&
+		    result == ISC_R_NOTFOUND)
+		       signer = NULL;
+		else {
+			tkey_log("dns_tkey_processquery: query was not "
+				 "properly signed - rejecting");
+			result = DNS_R_FORMERR;
+			goto failure;
+		}
+	} else
+		signer = &tsigner;
 
 	ISC_LIST_INIT(namelist);
 
@@ -595,6 +706,12 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 				goto failure;
 			}
 		}
+		if (tctx->domain == NULL) {
+			tkey_log("dns_tkey_processquery: tkey-domain not set");
+			dns_message_takebuffer(msg, &buf);
+			result = DNS_R_REFUSED;
+			goto failure;
+		}
 		result = dns_name_concatenate(&prefix, tctx->domain,
 					      keyname, buf);
 		dns_message_takebuffer(msg, &buf);
@@ -613,28 +730,26 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 	else
 		keyname = qname;
 
-	if (!dns_name_equal(&tkeyin.algorithm, DNS_TSIG_HMACMD5_NAME)) {
-		tkey_log("dns_tkey_processquery: tkey modes other than "
-			 "hmac-md5 are not supported");
-		tkeyout.error = dns_tsigerror_badkey;
-		goto failure_with_tkey;
-	}
-
 	switch (tkeyin.mode) {
 		case DNS_TKEYMODE_DIFFIEHELLMAN:
-			RETERR(process_dhtkey(msg, &signer, keyname, &tkeyin,
+			tkeyout.error = dns_rcode_noerror;
+			RETERR(process_dhtkey(msg, signer, keyname, &tkeyin,
 					      tctx, &tkeyout, ring,
 					      &namelist));
+			break;
+		case DNS_TKEYMODE_GSSAPI:
 			tkeyout.error = dns_rcode_noerror;
+			RETERR(process_gsstkey(msg, signer, keyname, &tkeyin,
+					       tctx, &tkeyout, ring,
+					       &namelist));
 			break;
 		case DNS_TKEYMODE_DELETE:
-			RETERR(process_deletetkey(msg, &signer, keyname,
+			tkeyout.error = dns_rcode_noerror;
+			RETERR(process_deletetkey(msg, signer, keyname,
 						  &tkeyin, &tkeyout,
 						  ring, &namelist));
-			tkeyout.error = dns_rcode_noerror;
 			break;
 		case DNS_TKEYMODE_SERVERASSIGNED:
-		case DNS_TKEYMODE_GSSAPI:
 		case DNS_TKEYMODE_RESOLVERASSIGNED:
 			result = DNS_R_NOTIMP;
 			goto failure;
@@ -646,10 +761,11 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 	dns_rdata_freestruct(&tkeyin);
 
 	RETERR(dns_message_gettemprdata(msg, &rdata));
-	RETERR(isc_buffer_allocate(msg->mctx, &dynbuf, 128));
+	RETERR(isc_buffer_allocate(msg->mctx, &dynbuf, 384));
 	result = dns_rdata_fromstruct(rdata, tkeyout.common.rdclass,
 				      tkeyout.common.rdtype, &tkeyout, dynbuf);
 	dns_name_free(&tkeyout.algorithm, msg->mctx);
+	freealg = ISC_FALSE;
 	if (tkeyout.key != NULL)
 		isc_mem_put(msg->mctx, tkeyout.key, tkeyout.keylen);
 	if (tkeyout.other != NULL)
@@ -682,6 +798,8 @@ dns_tkey_processquery(dns_message_t *msg, dns_tkeyctx_t *tctx,
 		dns_message_puttemprdata(msg, &rdata);
 	if (dynbuf != NULL)
 		isc_buffer_free(&dynbuf);
+	if (!ISC_LIST_EMPTY(namelist))
+		free_namelist(msg, &namelist);
 	return (result);
 }
 
@@ -786,7 +904,7 @@ dns_tkey_builddhquery(dns_message_t *msg, dst_key_t *key, dns_name_t *name,
 	tkey.expire = now + lifetime;
 	tkey.mode = DNS_TKEYMODE_DIFFIEHELLMAN;
 	if (nonce != NULL)
-		isc_buffer_region(nonce, &r);
+		isc_buffer_usedregion(nonce, &r);
 	else {
 		r.base = isc_mem_get(msg->mctx, 0);
 		r.length = 0;
@@ -827,6 +945,51 @@ dns_tkey_builddhquery(dns_message_t *msg, dst_key_t *key, dns_name_t *name,
 	return (result);
 }
 
+isc_result_t
+dns_tkey_buildgssquery(dns_message_t *msg, dns_name_t *name,
+		       dns_name_t *gname, void *cred,
+		       isc_uint32_t lifetime, void **context)
+{
+	dns_rdata_tkey_t tkey;
+	isc_result_t result;
+	isc_stdtime_t now;
+	isc_buffer_t token;
+	unsigned char array[1024];
+
+	REQUIRE(msg != NULL);
+	REQUIRE(name != NULL);
+	REQUIRE(gname != NULL);
+	REQUIRE(context != NULL && *context == NULL);
+
+	isc_buffer_init(&token, array, sizeof(array));
+	result = dst_gssapi_initctx(gname, cred, NULL, &token, context);
+	if (result != DNS_R_CONTINUE && result != ISC_R_SUCCESS)
+		return (result);
+
+	tkey.common.rdclass = dns_rdataclass_any;
+	tkey.common.rdtype = dns_rdatatype_tkey;
+	ISC_LINK_INIT(&tkey.common, link);
+	tkey.mctx = NULL;
+	dns_name_init(&tkey.algorithm, NULL);
+	dns_name_clone(DNS_TSIG_GSSAPI_NAME, &tkey.algorithm);
+	isc_stdtime_get(&now);
+	tkey.inception = now;
+	tkey.expire = now + lifetime;
+	tkey.mode = DNS_TKEYMODE_GSSAPI;
+	tkey.error = 0;
+	tkey.key = isc_buffer_base(&token);
+	tkey.keylen = isc_buffer_usedlength(&token);
+	tkey.other = NULL;
+	tkey.otherlen = 0;
+
+	RETERR(buildquery(msg, name, &tkey));
+
+	return (ISC_R_SUCCESS);
+
+ failure:
+	return (result);
+}
+
 isc_result_t
 dns_tkey_builddeletequery(dns_message_t *msg, dns_tsigkey_t *key) {
 	dns_rdata_tkey_t tkey;
@@ -839,7 +1002,7 @@ dns_tkey_builddeletequery(dns_message_t *msg, dns_tsigkey_t *key) {
 	ISC_LINK_INIT(&tkey.common, link);
 	tkey.mctx = msg->mctx;
 	dns_name_init(&tkey.algorithm, NULL);
-	dns_name_clone(&key->algorithm, &tkey.algorithm);
+	dns_name_clone(key->algorithm, &tkey.algorithm);
 	tkey.inception = tkey.expire = 0;
 	tkey.mode = DNS_TKEYMODE_DELETE;
 	tkey.error = 0;
@@ -882,11 +1045,11 @@ dns_tkey_processdhresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 			   dst_key_t *key, isc_buffer_t *nonce,
 			   dns_tsigkey_t **outkey, dns_tsig_keyring_t *ring)
 {
-	dns_rdata_t qtkeyrdata, rtkeyrdata;
+	dns_rdata_t qtkeyrdata = DNS_RDATA_INIT, rtkeyrdata = DNS_RDATA_INIT;
 	dns_name_t keyname, *tkeyname, *theirkeyname, *ourkeyname, *tempname;
 	dns_rdataset_t *theirkeyset = NULL, *ourkeyset = NULL;
-	dns_rdata_t theirkeyrdata;
-	dst_key_t *theirkey;
+	dns_rdata_t theirkeyrdata = DNS_RDATA_INIT;
+	dst_key_t *theirkey = NULL;
 	dns_rdata_tkey_t qtkey, rtkey;
 	unsigned char secretdata[256];
 	unsigned int sharedsize;
@@ -902,17 +1065,16 @@ dns_tkey_processdhresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 	REQUIRE(dst_key_isprivate(key));
 	if (outkey != NULL)
 		REQUIRE(*outkey == NULL);
-	REQUIRE(ring != NULL);
 
 	if (rmsg->rcode != dns_rcode_noerror)
 		return (ISC_RESULTCLASS_DNSRCODE + rmsg->rcode);
 	RETERR(find_tkey(rmsg, &tkeyname, &rtkeyrdata, DNS_SECTION_ANSWER));
-	RETERR(dns_rdata_tostruct(&rtkeyrdata, &rtkey, rmsg->mctx));
+	RETERR(dns_rdata_tostruct(&rtkeyrdata, &rtkey, NULL));
 	freertkey = ISC_TRUE;
 
 	RETERR(find_tkey(qmsg, &tempname, &qtkeyrdata,
 			 DNS_SECTION_ADDITIONAL));
-	RETERR(dns_rdata_tostruct(&qtkeyrdata, &qtkey, qmsg->mctx));
+	RETERR(dns_rdata_tostruct(&qtkeyrdata, &qtkey, NULL));
 
 	if (rtkey.error != dns_rcode_noerror ||
 	    rtkey.mode != DNS_TKEYMODE_DIFFIEHELLMAN ||
@@ -953,19 +1115,17 @@ dns_tkey_processdhresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 			break;
 		}
  next:
-                result = dns_message_nextname(rmsg, DNS_SECTION_ANSWER);
-        }
+		result = dns_message_nextname(rmsg, DNS_SECTION_ANSWER);
+	}
 
 	if (theirkeyset == NULL) {
 		tkey_log("dns_tkey_processdhresponse: failed to find server "
 			 "key");
-		result = DNS_R_INVALIDTKEY;
 		result = ISC_R_NOTFOUND;
 		goto failure;
 	}
 
 	dns_rdataset_current(theirkeyset, &theirkeyrdata);
-	theirkey = NULL;
 	RETERR(dns_dnssec_keyfromrdata(theirkeyname, &theirkeyrdata,
 				       rmsg->mctx, &theirkey));
 
@@ -979,7 +1139,7 @@ dns_tkey_processdhresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 	r.base = rtkey.key;
 	r.length = rtkey.keylen;
 	if (nonce != NULL)
-		isc_buffer_region(nonce, &r2);
+		isc_buffer_usedregion(nonce, &r2);
 	else {
 		r2.base = isc_mem_get(rmsg->mctx, 0);
 		r2.length = 0;
@@ -1011,11 +1171,71 @@ dns_tkey_processdhresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 	return (result);
 }
 
+isc_result_t
+dns_tkey_processgssresponse(dns_message_t *qmsg, dns_message_t *rmsg,
+			    dns_name_t *gname, void *cred, void **context,
+			    dns_tsigkey_t **outkey, dns_tsig_keyring_t *ring)
+{
+	dns_rdata_t rtkeyrdata = DNS_RDATA_INIT, qtkeyrdata = DNS_RDATA_INIT;
+	dns_name_t *tkeyname;
+	dns_rdata_tkey_t rtkey, qtkey;
+	isc_buffer_t outtoken;
+	dst_key_t *dstkey = NULL;
+	isc_region_t r;
+	isc_result_t result;
+	unsigned char array[1024];
+
+	REQUIRE(qmsg != NULL);
+	REQUIRE(rmsg != NULL);
+	REQUIRE(gname != NULL);
+	if (outkey != NULL)
+		REQUIRE(*outkey == NULL);
+
+	if (rmsg->rcode != dns_rcode_noerror)
+		return (ISC_RESULTCLASS_DNSRCODE + rmsg->rcode);
+	RETERR(find_tkey(rmsg, &tkeyname, &rtkeyrdata, DNS_SECTION_ANSWER));
+	RETERR(dns_rdata_tostruct(&rtkeyrdata, &rtkey, NULL));
+
+	RETERR(find_tkey(qmsg, &tkeyname, &qtkeyrdata,
+			 DNS_SECTION_ADDITIONAL));
+	RETERR(dns_rdata_tostruct(&qtkeyrdata, &qtkey, NULL));
+
+	if (rtkey.error != dns_rcode_noerror ||
+	    rtkey.mode != DNS_TKEYMODE_GSSAPI ||
+	    !dns_name_equal(&rtkey.algorithm, &rtkey.algorithm))
+	{
+		tkey_log("dns_tkey_processdhresponse: tkey mode invalid "
+			 "or error set");
+		result = DNS_R_INVALIDTKEY;
+		goto failure;
+	}
+
+	isc_buffer_init(&outtoken, array, sizeof(array));
+	r.base = rtkey.key;
+	r.length = rtkey.keylen;
+	RETERR(dst_gssapi_initctx(gname, cred, &r, &outtoken, context));
+
+	dstkey = NULL;
+	RETERR(dst_key_fromgssapi(dns_rootname, *context, rmsg->mctx,
+				  &dstkey));
+
+	RETERR(dns_tsigkey_createfromkey(tkeyname, DNS_TSIG_GSSAPI_NAME,
+					 dstkey, ISC_TRUE, NULL,
+					 rtkey.inception, rtkey.expire,
+					 rmsg->mctx, ring, outkey));
+
+	dns_rdata_freestruct(&rtkey);
+	return (result);
+
+ failure:
+	return (result);
+}
+
 isc_result_t
 dns_tkey_processdeleteresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 			       dns_tsig_keyring_t *ring)
 {
-	dns_rdata_t qtkeyrdata, rtkeyrdata;
+	dns_rdata_t qtkeyrdata = DNS_RDATA_INIT, rtkeyrdata = DNS_RDATA_INIT;
 	dns_name_t *tkeyname, *tempname;
 	dns_rdata_tkey_t qtkey, rtkey;
 	dns_tsigkey_t *tsigkey = NULL;
@@ -1028,11 +1248,11 @@ dns_tkey_processdeleteresponse(dns_message_t *qmsg, dns_message_t *rmsg,
 		return(ISC_RESULTCLASS_DNSRCODE + rmsg->rcode);
 
 	RETERR(find_tkey(rmsg, &tkeyname, &rtkeyrdata, DNS_SECTION_ANSWER));
-	RETERR(dns_rdata_tostruct(&rtkeyrdata, &rtkey, rmsg->mctx));
+	RETERR(dns_rdata_tostruct(&rtkeyrdata, &rtkey, NULL));
 
 	RETERR(find_tkey(qmsg, &tempname, &qtkeyrdata,
 			 DNS_SECTION_ADDITIONAL));
-	RETERR(dns_rdata_tostruct(&qtkeyrdata, &qtkey, qmsg->mctx));
+	RETERR(dns_rdata_tostruct(&qtkeyrdata, &qtkey, NULL));
 
 	if (rtkey.error != dns_rcode_noerror ||
 	    rtkey.mode != DNS_TKEYMODE_DELETE ||
